import numpy as np
from math import log
from itertools import compress

from imbcobra import classifier_cobra as cobra
from imbcobra import edited_knn, near_miss_v1, near_miss_v2, near_miss_v3, tomek_link, condensed_knn, knn_und

import logging


class CobraBoost:
    def __init__(self, X, y, machines, undersampling_method):
        self.X = X
        self.y = y

        self.model = cobra.CobraClassifier(machines = machines)
        self.majority_class_label = int(sum(y) > 0.5 * len(y))
        self.undersampling_method = undersampling_method

        self.weight_update = 0
        self.init_w = 1.0 / len(self.X)
        self.weight = np.full(len(self.X), self.init_w)
        

    def learn_parameters(self, iterations):
        verdict = self.undersampling_method.undersample(self.X, self.y, self.majority_class_label)
        X_undersampled, y_undersampled = self.X[verdict, :], self.y[verdict]
    
        for t in range(iterations):
            print("[Info]: Executing the iteration - {} of CobraBoost".format(t + 1))
            
            self.model.fit(X_undersampled, y_undersampled, sample_weight = self.weight[verdict])
        
            flag = self.y != self.model.predict(self.X)
            loss = sum(list(compress(self.weight, flag)))
            
            alpha = loss / (1 - loss)  

            if alpha <= 0:
                alpha = 0.0000001
            else:
                try:
                    alpha_hat = 0.5 * (np.log(1 - loss) - np.log(loss))
                except:
                    alpha_hat = 0
                    
                self.weight = self.weight * np.exp(-alpha_hat * self.y * self.model.predict(self.X))
                self.weight = self.weight / self.weight.sum()

            self.weight_update = alpha
            

    def predict(self, test_data):
        n = len(test_data)
        predicted_labels = np.zeros(n)

        for i in range(n):
            positive_score, negative_score = 0, 0
            
            if self.model.predict(test_data[i].reshape(1, -1)) == 1:
                positive_score += log(1/self.weight_update)
            else:
                negative_score += log(1/self.weight_update)
            
            if negative_score <= positive_score:
                predicted_labels[i] = 1

        return predicted_labels


import numpy as np

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import StratifiedKFold
from sklearn import metrics

import imbcobra.classifier_cobra
from imbcobra import near_miss_v1, near_miss_v2, near_miss_v3, knn_und, edited_knn, condensed_knn, tomek_link


def execute_model(X, y, num_splits, seed, model, with_undersampling = False, majority_class = 0, undersampling_method = knn_und):
  K_folds = StratifiedKFold(n_splits = num_splits, shuffle = True, random_state = seed)
  metrics_list = []
  
  # Feature Scaling
  sc = StandardScaler()
  X = sc.fit_transform(X)

  iterations = 1
  for train_idx, test_idx in K_folds.split(X, y):
    print("\n****************  Executing iteration - {} of KFold Data split  ****************".format(iterations))

    X_train, X_test = X[train_idx], X[test_idx]
    y_train, y_test = y[train_idx], y[test_idx]
    
    # Execute Undersampling on training data
    if with_undersampling == True:
      print("[Testing]: Count of test data before Undersampling = ", X_train.shape[0])
      verdict = undersampling_method.undersample(X_train, y_train, majority_class)

      X_train = X_train[verdict, :]
      y_train = y_train[verdict]

      # In-buit near miss algorithm
      # nr = NearMiss()
      # X_train, y_train = nr.fit_sample(X_train, y_train)

      # Note: Be careful while plotting, make sure same features are being compared
      # plt.scatter(X_train[:, 0], X_train[:, 1], marker = '.', c = y_train)
      # plt.show()

      print("[Testing]: Count of test data after Undersampling = ", X_train.shape[0])

    # Model Fitting & Predictions on test dataset
    y_pred = model(X_train, y_train, X_test)

    # Evaluation Metrics
    accuracy = metrics.accuracy_score(y_test, y_pred)
    precision, recall, F1_score, _ = metrics.precision_recall_fscore_support(y_test, y_pred, beta = 1.0, average = 'macro')
    metrics_list.append([accuracy, precision, recall, F1_score])

    iterations += 1
  
  metrics_list = np.mean(metrics_list, axis = 0)

  print("\n---------------  Cross-validated Evaluation Metrics  ---------------\n")
  print("Accuracy \t= \t", metrics_list[0])
  print("Precision \t= \t", metrics_list[1])
  print("Recall \t\t= \t", metrics_list[2])
  print("F1 score \t= \t", 2 * metrics_list[1] * metrics_list[2] / (metrics_list[1] + metrics_list[2]))


import numpy as np
from sklearn.neighbors import NearestNeighbors


def undersample(X, y, majority_class, ratio = 3.5, knn_algorithm = 'auto', knn_metric = 'euclidean'):
  n, _ = X.shape
  verdict = np.ones(n, dtype=bool)

  # Step - 1: Find k farthest neighbors (in minority class) of all the data points
  X_minority = []
  for idx, _ in enumerate(X):
    if y[idx] != majority_class:
      X_minority.append(X[idx])
      
  distances = []
  for idx, maj_val in enumerate(X):
    if idx % 500 == 0:
      print("[Executing]: Processing data point - {} while undersampling...".format(idx + 1))

    if y[idx] != majority_class:
      distances.append([0] * len(X_minority))    # To allow executing 'np.mean(distances, axis = 1)'
    else:
      current_distances = []
      for min_val in X_minority:
        current_distances.append(np.linalg.norm(maj_val - min_val))
      current_distances.sort(reverse = True)

      distances.append(current_distances)
  
  # Step - 2: Compute the average distance
  distances = np.mean(distances, axis = 1)
  
  # Step - 3: Select the points from majority class depending on the 'ratio' value
  X_majority_info = []  # stores a 2D value, where val at idx1 is the index in the original data and val at idx2 is the mean distance
  for i in range(n):
    class_label = y[i]
    if class_label == majority_class:
      X_majority_info.append([i, distances[i]])

  majority_size = len(X_majority_info)
  X_majority_info.sort(key=lambda x: x[1])
  req_count = min(majority_size, (int)(ratio * len(X_minority)))
  
  print("[Testing]: Count of majority samples after undersampling vs Count of minority samples = ", req_count, "vs", len(X_minority), "or", req_count * 100/len(X_minority), "%")
  for i in range(req_count, majority_size):
    verdict[X_majority_info[i][0]] = False  

  return verdict


import numpy as np
from sklearn.neighbors import NearestNeighbors


def undersample(X, y, majority_class, k = 5, knn_algorithm = 'auto', knn_metric = 'euclidean'):
  n, _ = X.shape
  verdict = np.ones(n, dtype=bool)

  # Step - 1: Find k nearest neighbors of all the data points 
  nearest_neighbors = NearestNeighbors(n_neighbors = k, algorithm = knn_algorithm, metric = knn_metric).fit(X)
  _, nearest_neighbors_idx = nearest_neighbors.kneighbors(X)

  # Step - 2: Delete elements from Majority Class
  for i in range(n):
    if(y[i] == majority_class):
      tmp = 0
      for j in nearest_neighbors_idx[i]:
        tmp += y[j]
      tmp/=3
      if(tmp<0.6):
        verdict[i] = False
    else:
      tmp = 0
      for j in nearest_neighbors_idx[i]:
        tmp += y[j]
      tmp/=3
      if(tmp>0.6):    
        for j in nearest_neighbors_idx[i]:
          if(y[j] == majority_class):
            verdict[j] == False

  return verdict            


import numpy as np
from sklearn.neighbors import NearestNeighbors


def undersample(X, y, majority_class, k = 3, ratio = 4.5, knn_algorithm = 'auto', knn_metric = 'euclidean'):
  n, _ = X.shape
  verdict = np.ones(n, dtype=bool)

  # Step - 1: Find k nearest neighbors (in minority class) of all the data points
  X_minority = []
  for idx, _ in enumerate(X):
    if y[idx] != majority_class:
      X_minority.append(X[idx])
      
  nearest_neighbors = NearestNeighbors(n_neighbors = k, algorithm = knn_algorithm, metric = knn_metric).fit(X_minority)
  distances, _ = nearest_neighbors.kneighbors(X)
  
  # Step - 2: Compute the average distance
  distances = np.mean(distances, axis = 1)
  
  # Step - 3: Select the points from majority class depending on the 'ratio' value
  X_majority_info = []  # stores a 2D value, where val at idx1 is the index in the original data and val at idx2 is the mean distance
  for i in range(n):
    class_label = y[i]
    if class_label == majority_class:
      X_majority_info.append([i, distances[i]])

  majority_size = len(X_majority_info)
  X_majority_info.sort(key=lambda x: x[1])
  req_count = min(majority_size, (int)(ratio * len(X_minority)))
  
  print("[Testing]: Count of majority samples after undersampling vs Count of minority samples = ", req_count, "vs", len(X_minority), "or", req_count * 100/len(X_minority), "%")
  for i in range(req_count, majority_size):
    verdict[X_majority_info[i][0]] = False  

  return verdict


import numpy as np
from sklearn.neighbors import NearestNeighbors


def undersample(X, y, majority_class, k = 50, t = 2, knn_algorithm = 'auto', knn_metric = 'euclidean'):
  n, _ = X.shape
  verdict = np.ones(n, dtype=bool)

  # Step - 1: Find k nearest neighbors of all the data points 
  nearest_neighbors = NearestNeighbors(n_neighbors = k, algorithm = knn_algorithm, metric = knn_metric).fit(X)
  _, nearest_neighbors_idx = nearest_neighbors.kneighbors(X)

  for i in range(n):
    # Step - 2: Identify the class label of X_i
    class_label = y[i]

    # Step - 3: Proceed only with majority class data points
    if class_label != majority_class:
      continue

    # Step - 4: Identify the count of minority class neighbors for the X_i
    minority_class_neighbors_count = 0
    neighbors_class_label = y[nearest_neighbors_idx[i]]

    for j in neighbors_class_label:
      if j != majority_class:
        minority_class_neighbors_count += 1

    # Step - 5: Mark the data point as False if count of minority class neighbors >= t
    if minority_class_neighbors_count >= t:
      verdict[i] = False

  return verdict


from distutils.core import setup
setup(
  name = 'imbcobra',
  packages = ['imbcobra'],
  version = '1.4',
  license = 'MIT', 
  description = 'COBRA for classification tasks (on Imbalanced Data)',
  author = ['Dr. Arabin Kumar Dey', 'Vishisht Priyadarshi', 'Aadi Gupta', 'Tejus Singla', 'Shashank Goyal'],
  author_email = 'vishishtpriyadarshi867@gmail.com',
  url = 'https://github.com/vishishtpriyadarshi/MA691-COBRA-6',
  download_url = 'https://github.com/vishishtpriyadarshi/MA691-COBRA-6/archive/refs/tags/v2.1.tar.gz',
  keywords = ['Classification', 'Imbalanced Data', 'Machine Learning'],
  install_requires=[
          'numpy',
          'scipy',
          'scikit-learn',
          'matplotlib',
          'seaborn',
          'pandas'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Science/Research',      
    'Topic :: Scientific/Engineering',
    'License :: OSI Approved :: MIT License',     
    'Programming Language :: Python :: 3.4',
  ],
)


import numpy as np
from sklearn.neighbors import NearestNeighbors


def undersample(X, y, majority_class, knn_algorithm = 'auto', knn_metric = 'euclidean'):
  n, _ = X.shape
  verdict = np.zeros(n, dtype=bool)

  # Step - 1: Find k nearest neighbors (in minority class) of all the data points
  store, store_y, grabstore, grabstore_y, s_idx, g_idx = [], [], [], [], [], []
  X_minority, X_majority = [], []

  for idx, val in enumerate(X):
    if y[idx] != majority_class:
      X_minority.append(X[idx])
      s_idx.append(idx)
      store.append(X[idx])
      store_y.append(y[idx])
    else:
      X_majority.append(X[idx])
      g_idx.append(idx)
      grabstore.append(X[idx])
      grabstore_y.append(y[idx])
    
  # Step - 2: Compute the average distance
  cnt = 100
  while((cnt != 0) and (len(grabstore)!=0)):
    cnt = 0
    i = 0
    for sn, i in enumerate(np.array(grabstore)):
      tmp = []
      nearest_neighbors = NearestNeighbors(n_neighbors = 1, algorithm = knn_algorithm, metric = knn_metric).fit(store)
      distances, nearest_neighbors_idx = nearest_neighbors.kneighbors(i.reshape(1,-1))
      if(store_y[nearest_neighbors_idx[0][0]] != majority_class):
        store.append(i)
        store_y.append(grabstore_y[sn])
        s_idx.append(g_idx[sn])
        tmp.append(sn)
        cnt += 1
    for z in tmp:
      del(grabstore[z])
      del(grabstore_y[z])
      del(g_idx[z])   
  
  for i in s_idx:
    verdict[i] = True

  return verdict